@@ -4,26 +4,32 @@ Standard ChatGPT
 from __future__ import annotations
 
 import base64
+import contextlib
 import json
 import logging
 import os
 import os.path as osp
+import sys
 import time
 import uuid
 from functools import wraps
 from os import environ
 from os import getenv
+from typing import NoReturn
 
 import requests
 from httpx import AsyncClient
 from OpenAIAuth import Authenticator
 from OpenAIAuth import Error as AuthError
 
-from .utils import get_input, create_session
+from .utils import create_completer
+from .utils import create_session
+from .utils import get_input
 
-logging.basicConfig(
-    format="%(asctime)s - %(name)s - %(levelname)s - %(funcName)s - %(message)s",
-)
+if __name__ == "__main__":
+    logging.basicConfig(
+        format="%(asctime)s - %(name)s - %(levelname)s - %(funcName)s - %(message)s",
+    )
 
 log = logging.getLogger(__name__)
 
@@ -68,7 +74,19 @@ def logger(is_timed: bool):
     return decorator
 
 
-BASE_URL = environ.get("CHATGPT_BASE_URL") or "https://chatgpt.duti.tech/"
+BASE_URL = environ.get("CHATGPT_BASE_URL") or "https://gpt.pawan.krd/backend-api/"
+
+
+class ErrorType:
+    # define consts for the error codes
+    USER_ERROR = -1
+    UNKNOWN_ERROR = 0
+    SERVER_ERROR = 1
+    RATE_LIMIT_ERROR = 2
+    INVALID_REQUEST_ERROR = 3
+    EXPIRED_ACCESS_TOKEN_ERROR = 4
+    INVALID_ACCESS_TOKEN_ERROR = 5
+    PROHIBITED_CONCURRENT_QUERY_ERROR = 6
 
 
 class Error(Exception):
@@ -82,17 +100,54 @@ class Error(Exception):
     3: Invalid request error
     4: Expired access token error
     5: Invalid access token error
+    6: Prohibited concurrent query error
     """
 
     source: str
     message: str
     code: int
 
-    def __init__(self, source: str, message: str, code: int = 0):
+    def __init__(self, source: str, message: str, code: int = 0) -> None:
         self.source = source
         self.message = message
         self.code = code
-        # Code
+
+    def __str__(self) -> str:
+        return f"{self.source}: {self.message} (code: {self.code})"
+
+    def __repr__(self) -> str:
+        return f"{self.source}: {self.message} (code: {self.code})"
+
+
+class colors:
+    """
+    Colors for printing
+    """
+
+    HEADER = "\033[95m"
+    OKBLUE = "\033[94m"
+    OKCYAN = "\033[96m"
+    OKGREEN = "\033[92m"
+    WARNING = "\033[93m"
+    FAIL = "\033[91m"
+    ENDC = "\033[0m"
+    BOLD = "\033[1m"
+    UNDERLINE = "\033[4m"
+
+    def __init__(self) -> None:
+        if getenv("NO_COLOR"):
+            self.HEADER = ""
+            self.OKBLUE = ""
+            self.OKCYAN = ""
+            self.OKGREEN = ""
+            self.WARNING = ""
+            self.FAIL = ""
+            self.ENDC = ""
+            self.BOLD = ""
+            self.UNDERLINE = ""
+
+
+bcolors = colors()
 
 
 class Chatbot:
@@ -131,7 +186,7 @@ class Chatbot:
         """
         user_home = getenv("HOME")
         if user_home is None:
-            self.cache_path = ".chatgpt_cache.json"
+            self.cache_path = osp.join(os.getcwd(), ".chatgpt_cache.json")
         else:
             # mkdir ~/.config/revChatGPT
             if not osp.exists(osp.join(user_home, ".config")):
@@ -161,6 +216,10 @@ class Chatbot:
                 "https": config["proxy"],
             }
             if isinstance(self.session, AsyncClient):
+                proxies = {
+                    "http://": config["proxy"],
+                    "https://": config["proxy"],
+                }
                 self.session = AsyncClient(proxies=proxies)
             else:
                 self.session.proxies.update(proxies)
@@ -174,7 +233,7 @@ class Chatbot:
         self.__check_credentials()
 
     @logger(is_timed=True)
-    def __check_credentials(self):
+    def __check_credentials(self) -> None:
         """Check login info and perform login
 
         Any one of the following is sufficient for login. Multiple login info can be provided at the same time and they will be used in the order listed below.
@@ -190,9 +249,7 @@ class Chatbot:
             self.__set_access_token(self.config["access_token"])
         elif "session_token" in self.config:
             pass
-        elif "email" in self.config and "password" in self.config:
-            pass
-        else:
+        elif "email" not in self.config or "password" not in self.config:
             raise Exception("Insufficient login details provided!")
         if "access_token" not in self.config:
             try:
@@ -201,7 +258,7 @@ class Chatbot:
                 raise error
 
     @logger(is_timed=False)
-    def __set_access_token(self, access_token: str):
+    def __set_access_token(self, access_token: str) -> None:
         """Set access token in request header and self.config, then cache it to file.
 
         Args:
@@ -219,6 +276,11 @@ class Chatbot:
                 "Referer": "https://chat.openai.com/chat",
             },
         )
+        self.session.cookies.update(
+            {
+                "library": "revChatGPT",
+            },
+        )
 
         self.config["access_token"] = access_token
 
@@ -258,13 +320,13 @@ class Chatbot:
                 raise Error(
                     source="__get_cached_access_token",
                     message="Invalid access token",
-                    code=5,
+                    code=ErrorType.INVALID_ACCESS_TOKEN_ERROR,
                 ) from None
             except json.JSONDecodeError:
                 raise Error(
                     source="__get_cached_access_token",
                     message="Invalid access token",
-                    code=5,
+                    code=ErrorType.INVALID_ACCESS_TOKEN_ERROR,
                 ) from None
 
             exp = d_access_token.get("exp", None)
@@ -272,7 +334,7 @@ class Chatbot:
                 raise Error(
                     source="__get_cached_access_token",
                     message="Access token expired",
-                    code=4,
+                    code=ErrorType.EXPIRED_ACCESS_TOKEN_ERROR,
                 )
 
         return access_token
@@ -293,7 +355,7 @@ class Chatbot:
         self.__write_cache(cache)
 
     @logger(is_timed=False)
-    def __write_cache(self, info: dict):
+    def __write_cache(self, info: dict) -> None:
         """Write cache info to file
 
         Args:
@@ -315,7 +377,7 @@ class Chatbot:
         return cached
 
     @logger(is_timed=True)
-    def __login(self):
+    def __login(self) -> None:
         if (
             "email" not in self.config or "password" not in self.config
         ) and "session_token" not in self.config:
@@ -370,7 +432,11 @@ class Chatbot:
 
         if parent_id is not None and conversation_id is None:
             log.error("conversation_id must be set once parent_id is set")
-            raise Error("User", "conversation_id must be set once parent_id is set", -1)
+            raise Error(
+                source="User",
+                message="conversation_id must be set once parent_id is set",
+                code=ErrorType.USER_ERROR,
+            )
 
         if conversation_id is not None and conversation_id != self.conversation_id:
             log.debug("Updating to new conversation by setting parent_id to None")
@@ -389,13 +455,11 @@ class Chatbot:
                         "Conversation ID %s not found in conversation mapping, try to get conversation history for the given ID",
                         conversation_id,
                     )
-                    try:
+                    with contextlib.suppress(Exception):
                         history = self.get_msg_history(conversation_id)
                         self.conversation_mapping[conversation_id] = history[
                             "current_node"
                         ]
-                    except Exception:
-                        pass
                 else:
                     log.debug(
                         "Conversation ID %s not found in conversation mapping, mapping conversations",
@@ -425,9 +489,9 @@ class Chatbot:
             ],
             "conversation_id": conversation_id,
             "parent_message_id": parent_id,
-            "model": "text-davinci-002-render-sha"
-            if not self.config.get("paid")
-            else "text-davinci-002-render-paid",
+            "model": "text-davinci-002-render-paid"
+            if self.config.get("paid")
+            else "text-davinci-002-render-sha",
         }
         log.debug("Sending the payload")
         log.debug(json.dumps(data, indent=2))
@@ -437,17 +501,22 @@ class Chatbot:
         )
         self.parent_id_prev_queue.append(data["parent_message_id"])
         response = self.session.post(
-            url=BASE_URL + "api/conversation",
+            url=f"{BASE_URL}conversation",
             data=json.dumps(data),
             timeout=timeout,
             stream=True,
         )
         self.__check_response(response)
         for line in response.iter_lines():
+            # remove b' and ' at the beginning and end and ignore case
             line = str(line)[2:-1]
-            if line == "Internal Server Error":
+            if line.lower() == "internal server error":
                 log.error("Internal Server Error: %s", line)
-                raise Exception("Error: " + str(line))
+                raise Error(
+                    source="ask",
+                    message="Internal Server Error",
+                    code=ErrorType.SERVER_ERROR,
+                )
             if line == "" or line is None:
                 continue
             if "data: " in line:
@@ -465,21 +534,55 @@ class Chatbot:
                 continue
             if not self.__check_fields(line):
                 log.error("Field missing", exc_info=True)
-                if (
-                    line.get("detail")
-                    == "Too many requests in 1 hour. Try again later."
-                ):
-                    log.error("Rate limit exceeded")
-                    raise Error(source="ask", message=line.get("detail"), code=2)
-                if line.get("detail", {}).get("code") == "invalid_api_key":
-                    log.error("Invalid access token")
-                    raise Error(
-                        source="ask",
-                        message=line.get("detail", {}).get("message"),
-                        code=3,
-                    )
+                line_detail = line.get("detail")
+                if isinstance(line_detail, str):
+                    if (
+                        line_detail.lower()
+                        == "too many requests in 1 hour. try again later."
+                    ):
+                        log.error("Rate limit exceeded")
+                        raise Error(
+                            source="ask",
+                            message=line.get("detail"),
+                            code=ErrorType.RATE_LIMIT_ERROR,
+                        )
+                    if line_detail.lower().startswith(
+                        "only one message at a time.",
+                    ):
+                        log.error("Prohibited concurrent query")
+                        raise Error(
+                            source="ask",
+                            message=line_detail,
+                            code=ErrorType.PROHIBITED_CONCURRENT_QUERY_ERROR,
+                        )
+                    if line_detail.lower() == "invalid_api_key":
+                        log.error("Invalid access token")
+                        raise Error(
+                            source="ask",
+                            message=line_detail,
+                            code=ErrorType.INVALID_REQUEST_ERROR,
+                        )
+                    if line_detail.lower() == "invalid_token":
+                        log.error("Invalid access token")
+                        raise Error(
+                            source="ask",
+                            message=line_detail,
+                            code=ErrorType.INVALID_ACCESS_TOKEN_ERROR,
+                        )
+                elif isinstance(line_detail, dict):
+                    if line_detail.get("code") == "invalid_jwt":
+                        log.error("Invalid access token")
+                        raise Error(
+                            source="ask",
+                            message=line_detail.get("message", "invalid_jwt"),
+                            code=ErrorType.INVALID_ACCESS_TOKEN_ERROR,
+                        )
 
-                raise Error(source="ask", message="Field missing", code=1)
+                raise Error(
+                    source="ask",
+                    message="Field missing",
+                    code=ErrorType.SERVER_ERROR,
+                )
             message = line["message"]["content"]["parts"][0]
             if message == prompt:
                 continue
@@ -508,14 +611,12 @@ class Chatbot:
     def __check_fields(self, data: dict) -> bool:
         try:
             data["message"]["content"]
-        except TypeError:
-            return False
-        except KeyError:
+        except (TypeError, KeyError):
             return False
         return True
 
     @logger(is_timed=False)
-    def __check_response(self, response):
+    def __check_response(self, response: requests.Response) -> None:
         """Make sure response is success
 
         Args:
@@ -526,7 +627,11 @@ class Chatbot:
         """
         if response.status_code != 200:
             print(response.text)
-            raise Error("OpenAI", response.status_code, response.text)
+            raise Error(
+                source="OpenAI",
+                message=response.text,
+                code=response.status_code,
+            )
 
     @logger(is_timed=True)
     def get_conversations(
@@ -540,7 +645,7 @@ class Chatbot:
         :param offset: Integer
         :param limit: Integer
         """
-        url = BASE_URL + f"api/conversations?offset={offset}&limit={limit}"
+        url = f"{BASE_URL}conversations?offset={offset}&limit={limit}"
         response = self.session.get(url)
         self.__check_response(response)
         if encoding is not None:
@@ -555,21 +660,20 @@ class Chatbot:
         :param id: UUID of conversation
         :param encoding: String
         """
-        url = BASE_URL + f"api/conversation/{convo_id}"
+        url = f"{BASE_URL}conversation/{convo_id}"
         response = self.session.get(url)
         self.__check_response(response)
         if encoding is not None:
             response.encoding = encoding
-        data = json.loads(response.text)
-        return data
+        return json.loads(response.text)
 
     @logger(is_timed=True)
-    def gen_title(self, convo_id: str, message_id: str):
+    def gen_title(self, convo_id: str, message_id: str) -> None:
         """
         Generate title for conversation
         """
         response = self.session.post(
-            BASE_URL + f"api/conversation/gen_title/{convo_id}",
+            f"{BASE_URL}conversation/gen_title/{convo_id}",
             data=json.dumps(
                 {"message_id": message_id, "model": "text-davinci-002-render"},
             ),
@@ -577,37 +681,37 @@ class Chatbot:
         self.__check_response(response)
 
     @logger(is_timed=True)
-    def change_title(self, convo_id: str, title: str):
+    def change_title(self, convo_id: str, title: str) -> None:
         """
         Change title of conversation
         :param id: UUID of conversation
         :param title: String
         """
-        url = BASE_URL + f"api/conversation/{convo_id}"
+        url = f"{BASE_URL}conversation/{convo_id}"
         response = self.session.patch(url, data=json.dumps({"title": title}))
         self.__check_response(response)
 
     @logger(is_timed=True)
-    def delete_conversation(self, convo_id: str):
+    def delete_conversation(self, convo_id: str) -> None:
         """
         Delete conversation
         :param id: UUID of conversation
         """
-        url = BASE_URL + f"api/conversation/{convo_id}"
+        url = f"{BASE_URL}conversation/{convo_id}"
         response = self.session.patch(url, data='{"is_visible": false}')
         self.__check_response(response)
 
     @logger(is_timed=True)
-    def clear_conversations(self):
+    def clear_conversations(self) -> None:
         """
         Delete all conversations
         """
-        url = BASE_URL + "api/conversations"
+        url = f"{BASE_URL}conversations"
         response = self.session.patch(url, data='{"is_visible": false}')
         self.__check_response(response)
 
     @logger(is_timed=False)
-    def __map_conversations(self):
+    def __map_conversations(self) -> None:
         conversations = self.get_conversations()
         histories = [self.get_msg_history(x["id"]) for x in conversations]
         for x, y in zip(conversations, histories):
@@ -664,7 +768,11 @@ class AsyncChatbot(Chatbot):
         Ask a question to the chatbot
         """
         if parent_id is not None and conversation_id is None:
-            raise Error("User", "conversation_id must be set once parent_id is set", 1)
+            raise Error(
+                source="User",
+                message="conversation_id must be set once parent_id is set",
+                code=ErrorType.SERVER_ERROR,
+            )
 
         if conversation_id is not None and conversation_id != self.conversation_id:
             self.parent_id = None
@@ -689,9 +797,9 @@ class AsyncChatbot(Chatbot):
             ],
             "conversation_id": conversation_id,
             "parent_message_id": parent_id,
-            "model": "text-davinci-002-render-sha"
-            if not self.config.get("paid")
-            else "text-davinci-002-render-paid",
+            "model": "text-davinci-002-render-paid"
+            if self.config.get("paid")
+            else "text-davinci-002-render-sha",
         }
 
         self.conversation_id_prev_queue.append(
@@ -701,7 +809,7 @@ class AsyncChatbot(Chatbot):
 
         async with self.session.stream(
             method="POST",
-            url=BASE_URL + "api/conversation",
+            url=f"{BASE_URL}conversation",
             data=json.dumps(data),
             timeout=timeout,
         ) as response:
@@ -719,7 +827,7 @@ class AsyncChatbot(Chatbot):
                 except json.decoder.JSONDecodeError:
                     continue
                 if not self.__check_fields(line):
-                    raise Exception("Field missing. Details: " + str(line))
+                    raise Exception(f"Field missing. Details: {str(line)}")
 
                 message = line["message"]["content"]["parts"][0]
                 conversation_id = line["conversation_id"]
@@ -747,7 +855,7 @@ class AsyncChatbot(Chatbot):
         :param offset: Integer
         :param limit: Integer
         """
-        url = BASE_URL + f"api/conversations?offset={offset}&limit={limit}"
+        url = f"{BASE_URL}conversations?offset={offset}&limit={limit}"
         response = await self.session.get(url)
         self.__check_response(response)
         data = json.loads(response.text)
@@ -758,19 +866,18 @@ class AsyncChatbot(Chatbot):
         Get message history
         :param id: UUID of conversation
         """
-        url = BASE_URL + f"api/conversation/{convo_id}"
+        url = f"{BASE_URL}conversation/{convo_id}"
         response = await self.session.get(url)
         if encoding is not None:
             response.encoding = encoding
             self.__check_response(response)
-            data = json.loads(response.text)
-            return data
+            return json.loads(response.text)
 
-    async def gen_title(self, convo_id, message_id):
+    async def gen_title(self, convo_id: str, message_id: str) -> None:
         """
         Generate title for conversation
         """
-        url = BASE_URL + f"api/conversation/gen_title/{convo_id}"
+        url = f"{BASE_URL}conversation/gen_title/{convo_id}"
         response = await self.session.post(
             url,
             data=json.dumps(
@@ -779,34 +886,34 @@ class AsyncChatbot(Chatbot):
         )
         await self.__check_response(response)
 
-    async def change_title(self, convo_id, title):
+    async def change_title(self, convo_id: str, title: str) -> None:
         """
         Change title of conversation
         :param convo_id: UUID of conversation
         :param title: String
         """
-        url = BASE_URL + f"api/conversation/{convo_id}"
+        url = f"{BASE_URL}conversation/{convo_id}"
         response = await self.session.patch(url, data=f'{{"title": "{title}"}}')
         self.__check_response(response)
 
-    async def delete_conversation(self, convo_id):
+    async def delete_conversation(self, convo_id: str) -> None:
         """
         Delete conversation
         :param convo_id: UUID of conversation
         """
-        url = BASE_URL + f"api/conversation/{convo_id}"
+        url = f"{BASE_URL}conversation/{convo_id}"
         response = await self.session.patch(url, data='{"is_visible": false}')
         self.__check_response(response)
 
-    async def clear_conversations(self):
+    async def clear_conversations(self) -> None:
         """
         Delete all conversations
         """
-        url = BASE_URL + "api/conversations"
+        url = f"{BASE_URL}conversations"
         response = await self.session.patch(url, data='{"is_visible": false}')
         self.__check_response(response)
 
-    async def __map_conversations(self):
+    async def __map_conversations(self) -> None:
         conversations = await self.get_conversations()
         histories = [await self.get_msg_history(x["id"]) for x in conversations]
         for x, y in zip(conversations, histories):
@@ -815,13 +922,11 @@ class AsyncChatbot(Chatbot):
     def __check_fields(self, data: dict) -> bool:
         try:
             data["message"]["content"]
-        except TypeError:
-            return False
-        except KeyError:
+        except (TypeError, KeyError):
             return False
         return True
 
-    def __check_response(self, response):
+    def __check_response(self, response) -> None:
         response.raise_for_status()
 
 
@@ -834,15 +939,12 @@ def configure():
     Looks for a config file in the following locations:
     """
     config_files = ["config.json"]
-    xdg_config_home = getenv("XDG_CONFIG_HOME")
-    if xdg_config_home:
+    if xdg_config_home := getenv("XDG_CONFIG_HOME"):
         config_files.append(f"{xdg_config_home}/revChatGPT/config.json")
-    user_home = getenv("HOME")
-    if user_home:
+    if user_home := getenv("HOME"):
         config_files.append(f"{user_home}/.config/revChatGPT/config.json")
 
-    config_file = next((f for f in config_files if osp.exists(f)), None)
-    if config_file:
+    if config_file := next((f for f in config_files if osp.exists(f)), None):
         with open(config_file, encoding="utf-8") as f:
             config = json.load(f)
     else:
@@ -852,11 +954,10 @@ def configure():
 
 
 @logger(is_timed=False)
-def main(config: dict):
+def main(config: dict) -> NoReturn:
     """
     Main function for the chatGPT program.
     """
-    print("Logging in...")
     chatbot = Chatbot(
         config,
         conversation_id=config.get("conversation_id"),
@@ -904,25 +1005,36 @@ def main(config: dict):
                 )
                 print("Please include conversation UUID in command")
         elif command == "!exit":
-            exit(0)
+            sys.exit(0)
         else:
             return False
         return True
 
     session = create_session()
-    while True:
-        prompt = get_input("\nYou:\n", session=session)
-        if prompt.startswith("!"):
-            if handle_commands(prompt):
+    completer = create_completer(
+        ["!help", "!reset", "!config", "!rollback", "!exit", "!setconversation"],
+    )
+    print()
+    try:
+        while True:
+            print(f"{bcolors.OKBLUE + bcolors.BOLD}You: {bcolors.ENDC}")
+
+            prompt = get_input(session=session, completer=completer)
+            if prompt.startswith("!") and handle_commands(prompt):
                 continue
-        print()
-        print("Chatbot: ")
-        prev_text = ""
-        for data in chatbot.ask(prompt):
-            message = data["message"][len(prev_text) :]
-            print(message, end="", flush=True)
-            prev_text = data["message"]
-        print()
+
+            print()
+            print(f"{bcolors.OKGREEN + bcolors.BOLD}Chatbot: {bcolors.ENDC}")
+            prev_text = ""
+            for data in chatbot.ask(prompt):
+                message = data["message"][len(prev_text) :]
+                print(message, end="", flush=True)
+                prev_text = data["message"]
+            print(bcolors.ENDC)
+            print()
+    except (KeyboardInterrupt, EOFError):
+        print("Exiting...")
+        sys.exit(0)
 
 
 if __name__ == "__main__":
@@ -933,4 +1045,7 @@ if __name__ == "__main__":
         """,
     )
     print("Type '!help' to show a full list of commands")
+    print(
+        f"{bcolors.BOLD}{bcolors.WARNING}Press Esc followed by Enter or Alt+Enter to send a message.{bcolors.ENDC}",
+    )
     main(configure())
